import json
from abc import abstractmethod
from typing import Any, Optional

from .trie import AbstractTrie, Iterator, TrieKey


class SerializedTrie(AbstractTrie):
    @property
    @abstractmethod
    def _trie(self):
        pass

    def close(self):
        self._trie.close()

    def commit(self):
        self._trie.commit()

    def rollback(self):
        self._trie.rollback()

    @abstractmethod
    def _load(self, key: TrieKey, value: Optional[bytes]) -> Optional[Any]:
        pass

    @abstractmethod
    def _dump(self, key: TrieKey, value: Optional[Any]) -> Optional[bytes]:
        pass

    def __setitem__(self, key, value):
        self._trie[key] = self._dump(key, value)

    def __getitem__(self, key):
        raw = self._trie[key]
        return self._load(key, raw)

    def __delitem__(self, key):
        del self._trie[key]

    def __len__(self):
        return len(self._trie)

    def view(self, key: Optional[TrieKey] = None) -> "SerializedTrie":
        if not key:
            return self

        raw_trie = self._trie.view(key)
        trie = type(self)()
        # pylint: disable-next=protected-access
        trie._trie = raw_trie  # type: ignore
        return trie

    def items(self, *args, **kwargs):
        yield from (
            (key, self._load(key, raw))
            for key, raw in self._trie.items(*args, **kwargs)
        )

    def ls(self, key, with_values=False):
        entries = self._trie.ls(key, with_values=with_values)
        if with_values:
            yield from (
                (ekey, self._load(ekey, evalue)) for ekey, evalue in entries
            )
        else:
            yield from entries

    def traverse(self, node_factory, prefix=None):
        def _node_factory_wrapper(path_conv, path, children, value):
            return node_factory(
                path_conv, path, children, self._load(path, value)
            )

        return self._trie.traverse(_node_factory_wrapper, prefix=prefix)

    def diff(self, *args, **kwargs):
        yield from self._trie.diff(*args, **kwargs)

    def has_node(self, key):
        return self._trie.has_node(key)

    def shortest_prefix(self, key):
        sprefix = self._trie.shortest_prefix(key)
        if sprefix is None:
            return None

        skey, raw = sprefix
        return key, self._load(skey, raw)

    def prefixes(self, key):
        for prefix, raw in self._trie.prefixes(key):
            yield (prefix, self._load(prefix, raw))

    def longest_prefix(self, key):
        lprefix = self._trie.longest_prefix(key)
        if lprefix is None:
            return None

        lkey, raw = lprefix
        return lkey, self._load(lkey, raw)

    def __iter__(self) -> Iterator[TrieKey]:
        yield from self._trie


class JSONTrie(SerializedTrie):  # pylint: disable=abstract-method
    def _load(self, key: TrieKey, value: Optional[bytes]) -> Optional[Any]:
        if value is None:
            return None
        return json.loads(value.decode("utf-8"))

    def _dump(self, key: TrieKey, value: Optional[Any]) -> Optional[bytes]:
        if value is None:
            return None
        return json.dumps(value).encode("utf-8")



from abc import abstractmethod
from collections.abc import MutableMapping
from typing import Iterator, NamedTuple, Optional, Tuple, Union

from attrs import define


class ShortKeyError(KeyError):
    """Raised when given key is a prefix of an existing longer key
    but does not have a value associated with itself."""


TrieKey = Union[Tuple[()], Tuple[str]]
TrieStep = Tuple[Optional[TrieKey], Optional[bytes]]


class TrieNode(NamedTuple):
    key: TrieKey
    value: Optional[bytes]


ADD = "add"
MODIFY = "modify"
RENAME = "rename"
DELETE = "delete"
UNCHANGED = "unchanged"


@define(frozen=True, hash=True, order=True)
class Change:
    typ: str
    old: Optional[TrieNode]
    new: Optional[TrieNode]

    @property
    def key(self) -> TrieKey:
        if self.typ == RENAME:
            raise ValueError

        if self.typ == ADD:
            entry = self.new
        else:
            entry = self.old

        assert entry
        assert entry.key
        return entry.key

    def __bool__(self) -> bool:
        return self.typ != UNCHANGED


class AbstractTrie(MutableMapping):
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    @classmethod
    @abstractmethod
    def open(cls, path: str) -> "AbstractTrie":
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    @abstractmethod
    def commit(self) -> None:
        pass

    @abstractmethod
    def rollback(self) -> None:
        pass

    @abstractmethod
    def items(  # type: ignore
        self, prefix: Optional[TrieKey] = None, shallow: Optional[bool] = False
    ) -> Iterator[Tuple[TrieKey, bytes]]:
        pass

    @abstractmethod
    def view(self, key: Optional[TrieKey] = None) -> "AbstractTrie":
        pass

    @abstractmethod
    def has_node(self, key: TrieKey) -> bool:
        pass

    @abstractmethod
    def prefixes(self, key: TrieKey) -> Iterator[TrieStep]:
        pass

    @abstractmethod
    def shortest_prefix(self, key: TrieKey) -> Optional[TrieStep]:
        pass

    @abstractmethod
    def longest_prefix(self, key: TrieKey) -> Optional[TrieStep]:
        pass

    @abstractmethod
    # pylint: disable-next=invalid-name
    def ls(
        self, key: TrieKey, with_values: bool = False
    ) -> Iterator[Union[TrieKey, TrieNode]]:
        pass

    @abstractmethod
    def traverse(
        self, node_factory, prefix: Optional[TrieKey]
    ) -> Iterator[Tuple[TrieKey, bytes]]:
        pass

    @abstractmethod
    def diff(
        self, old: TrieKey, new: TrieKey, with_unchanged: bool = False
    ) -> Iterator[Change]:
        pass



import pygtrie

from .trie import (
    ADD,
    DELETE,
    MODIFY,
    UNCHANGED,
    AbstractTrie,
    Change,
    ShortKeyError,
    TrieNode,
)


class PyGTrie(AbstractTrie):
    def __init__(self, *args, **kwargs):
        self._trie = pygtrie.Trie()
        super().__init__(*args, **kwargs)

    @classmethod
    def open(cls, path):
        raise NotImplementedError

    def close(self):
        pass

    def commit(self):
        pass

    def rollback(self):
        raise NotImplementedError

    def __setitem__(self, key, value):
        self._trie[key] = value

    def __iter__(self):
        yield from (key for key, _ in self.items())

    def __getitem__(self, key):
        try:
            return self._trie[key]
        except pygtrie.ShortKeyError as exc:
            raise ShortKeyError(key) from exc

    def __delitem__(self, key):
        del self._trie[key]

    def __len__(self):
        return len(self._trie)

    def has_node(self, key):
        return self._trie.has_node(key)

    def items(self, prefix=None, shallow=False):
        kwargs = {"shallow": shallow}
        if prefix is not None:
            kwargs["prefix"] = prefix

        yield from self._trie.iteritems(**kwargs)

    def prefixes(self, key):
        return self._trie.prefixes(key)

    def shortest_prefix(self, key):
        return self._trie.shortest_prefix(key)

    def longest_prefix(self, key):
        ret = self._trie.longest_prefix(key)
        if ret == pygtrie.Trie._NONE_STEP:  # pylint: disable=protected-access
            return None
        return tuple(ret)

    def traverse(self, node_factory, prefix=None):
        kwargs = {}
        if prefix is not None:
            kwargs["prefix"] = prefix

        return self._trie.traverse(node_factory, **kwargs)

    def view(self, key=None):
        ret = PyGTrie()
        kwargs = {}
        if key is not None:
            kwargs["prefix"] = key

        for ikey, value in self._trie.iteritems(**kwargs):
            ret[ikey[len(key) :]] = value

        return ret

    def ls(self, key, with_values=False):
        def node_factory(_, nkey, children, value=None):
            if nkey == key:
                return children

            if with_values:
                return nkey, value

            return nkey

        return self.traverse(node_factory, prefix=key)

    def diff(self, old, new, with_unchanged=False):
        # FIXME this is not the most optimal implementation
        old_keys = {key for key, _ in self.items(old or ())}
        new_keys = {key for key, _ in self.items(new or ())}

        for key in old_keys | new_keys:
            old_entry = self.get(key)
            new_entry = self.get(key)

            typ = UNCHANGED
            if old_entry and not new_entry:
                typ = DELETE
            elif not old_entry and new_entry:
                typ = ADD
            elif old_entry != new_entry:
                typ = MODIFY
            elif not with_unchanged:
                continue

            old_node = (
                TrieNode(key[len(old) :], old_entry) if old_entry else None
            )
            new_node = (
                TrieNode(key[len(new) :], new_entry) if new_entry else None
            )
            yield Change(typ, old_node, new_node)



import pytest

from sqltrie import PyGTrie, SQLiteTrie

NFILES = 10000
NSUBDIRS = 3


@pytest.fixture(scope="session")
def items():
    ret = {}

    files = {str(idx): bytes(idx) for idx in range(NFILES)}
    for subdir in range(NSUBDIRS):
        ret[str(subdir)] = files.copy()

    return ret


@pytest.mark.parametrize("cls", [SQLiteTrie, PyGTrie])
def test_set(benchmark, items, cls):
    def _set():
        trie = cls()

        for subdir in range(NSUBDIRS):
            for idx in range(NFILES):
                trie[(str(subdir), str(idx))] = bytes(idx)

    benchmark(_set)


@pytest.mark.parametrize("cls", [SQLiteTrie, PyGTrie])
def test_items(benchmark, items, cls):
    trie = cls()

    for subdir in range(NSUBDIRS):
        for idx in range(NFILES):
            trie[(str(subdir), str(idx))] = bytes(idx)

    def _items():
        list(trie.items())

    benchmark(_items)


@pytest.mark.parametrize("cls", [SQLiteTrie, PyGTrie])
def test_ls(benchmark, items, cls):
    trie = cls()

    for subdir in range(NSUBDIRS):
        for idx in range(NFILES):
            trie[(str(subdir), str(idx))] = bytes(idx)

    def _ls():
        list(trie.ls(("1",)))

    benchmark(_ls)


@pytest.mark.parametrize("cls", [SQLiteTrie, PyGTrie])
def test_diff(benchmark, items, cls):
    trie = cls()

    for subdir in range(NSUBDIRS):
        for idx in range(NFILES):
            trie[(str(subdir), str(idx))] = bytes(idx)

    def _diff():
        list(trie.diff(None, None))

    benchmark(_diff)
